@@ -1,11 +1,14 @@
 ;;; desktop.el --- save partial status of Emacs when killed
 
-;; Copyright (C) 1993, 1994, 1995 Free Software Foundation, Inc.
+;; Copyright (C) 1993, 1994, 1995, 1997, 2000, 2001
+;;   Free Software Foundation, Inc.
 
 ;; Author: Morten Welinder <terra@diku.dk>
 ;; Version: 2.09?  RMS has an obnoxious tendency to remove version
-;; numbers from packages, and he did in this case.
-;; Keywords: customization
+;; numbers from packages, and he did in this case. (In any case,
+;; the primary source of this package is now within Emacs, so this
+;; version number is passing irrelevant.)
+;; Keywords: convenience
 ;; Favourite-brand-of-beer: None, I hate beer.
 
 ;; This file is part of XEmacs.
@@ -25,7 +28,7 @@
 ;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 ;; 02111-1307, USA.
 
-;;; Synched up with: FSF 19.34.
+;;; Synched up with: FSF 20.1.
 
 ;;; Commentary:
 
@@ -39,18 +42,16 @@
 ;;		- buffer-read-only
 ;;		- some local variables
 
-;; To use this, first put these three lines in the bottom of your .emacs
+;; To use this, first put these two lines in the bottom of your .emacs
 ;; file (the later the better):
 ;;
-;;	(load "desktop")
 ;;	(desktop-load-default)
 ;;	(desktop-read)
 ;;
-;; Between the second and the third line you may wish to add something that
-;; updates the variables `desktop-globals-to-save' and/or
-;; `desktop-locals-to-save'.  If for instance you want to save the local
-;; variable `foobar' for every buffer in which it is local, you could add
-;; the line
+;; Between these two lines you may wish to add something that updates the
+;; variables `desktop-globals-to-save' and/or `desktop-locals-to-save'.  If
+;; for instance you want to save the local variable `foobar' for every buffer
+;; in which it is local, you could add the line
 ;;
 ;;	(setq desktop-locals-to-save (cons 'foobar desktop-locals-to-save))
 ;;
@@ -63,8 +64,8 @@
 ;;		     (desktop-truncate regexp-search-ring 3)))
 ;;
 ;; which will make sure that no more than three search items are saved.  You
-;; must place this line *after* the (load "desktop") line.  See also the
-;; variable desktop-save-hook.
+;; must place this line *after* the `(desktop-load-default)' line.  See also
+;; the variable `desktop-save-hook'.
 
 ;; Start Emacs in the root directory of your "project". The desktop saver
 ;; is inactive by default.  You activate it by M-x desktop-save RET.  When
@@ -75,6 +76,14 @@
 ;; act as a default desktop when you start Emacs from a directory that
 ;; doesn't have its own.  I never do this, but you may want to.
 
+;; Some words on minor modes: Most minor modes are controlled by
+;; buffer-local variables, which have a standard save / restore
+;; mechanism.  To handle all minor modes, we take the following
+;; approach: (1) check whether the variable name from
+;; `minor-mode-alist' is also a function; and (2) use translation
+;; table `desktop-minor-mode-table' in the case where the two names
+;; are not the same.
+
 ;; By the way: don't use desktop.el to customize Emacs -- the file .emacs
 ;; in your home directory is used for that.  Saving global default values
 ;; for buffers is an example of misuse.
@@ -104,20 +113,30 @@
 (eval-when-compile
   ;; We use functions from these modules
   ;; We can't (require 'mh-e) since that wants to load something.
-  (require 'info))
+  (mapcar 'require '(info dired reporter)))
 ;; ----------------------------------------------------------------------------
 ;; USER OPTIONS -- settings you might want to play with.
 ;; ----------------------------------------------------------------------------
 
 (defgroup desktop nil
-  "Save partial status of Emacs when killed"
+  "Save partial status of Emacs when you exit."
   :group 'frames)
 
-(defconst desktop-basefilename
+(defcustom desktop-enable nil
+  "*Non-nil enable Desktop to save the state of Emacs when you exit."
+  :group 'desktop
+  :type 'boolean
+  :require 'desktop
+  :initialize 'custom-initialize-default
+  :version "20.3")
+
+(defcustom desktop-basefilename
   (if (or (eq system-type 'ms-dos) (eq system-type 'windows-nt))
       "emacs.dsk" ; Ms-Dos does not support multiple dots in file name
     ".emacs.desktop")
-  "File for Emacs desktop, not including the directory name.")
+  "File for Emacs desktop, not including the directory name."
+  :type 'file
+  :group 'desktop)
 
 (defcustom desktop-missing-file-warning t
   "*If non-nil then desktop warns when a file no longer exists.
@@ -129,8 +148,6 @@ Otherwise it simply ignores that file."
   (list 'desktop-missing-file-warning
 	;; Feature: saving kill-ring implies saving kill-ring-yank-pointer
 	;; 'kill-ring
-	'tags-file-name
-	'tags-table-list
 	'search-ring
 	'regexp-search-ring
 	'register-alist
@@ -166,33 +183,67 @@ The variables are saved only when they really are local."
 
 ;; We skip .log files because they are normally temporary.
 ;;         (ftp) files because they require passwords and whatnot.
-;;         TAGS files to save time (tags-file-name is saved instead).
+;;         TAGS files to save time.
 (defcustom desktop-buffers-not-to-save
  "\\(^nn\\.a[0-9]+\\|\\.log\\|(ftp)\\|^tags\\|^TAGS\\)$"
  "Regexp identifying buffers that are to be excluded from saving."
  :type 'regexp
  :group 'desktop)
 
-;; Skip ange-ftp files
+;; Skip efs files
 (defcustom desktop-files-not-to-save
   "^/[^/:]*:"
   "Regexp identifying files whose buffers are to be excluded from saving."
   :type 'regexp
   :group 'desktop)
 
-(defvar desktop-buffer-major-mode nil
-  "When desktop creates a buffer, this holds the desired Major mode.")
+(defcustom desktop-buffer-modes-to-save
+  '(Info-mode rmail-mode)
+  "If a buffer is of one of these major modes, save the buffer name.
+It is up to the functions in `desktop-buffer-handlers' to decide
+whether the buffer should be recreated or not, and how."
+  :type '(repeat symbol)
+  :group 'desktop)
 
-(defvar desktop-buffer-file-name nil
-  "When desktop creates a buffer, this holds the file name to visit.")
+(defcustom desktop-modes-not-to-save nil
+  "List of major modes whose buffers should not be saved."
+  :type '(repeat symbol)
+  :group 'desktop)
 
-(defvar desktop-buffer-name nil
-  "When desktop creates a buffer, this holds the desired buffer name.")
+(defcustom desktop-buffer-major-mode nil
+  "When desktop creates a buffer, this holds the desired Major mode."
+  :type 'symbol
+  :group 'desktop)
+
+(defcustom desktop-buffer-file-name nil
+  "When desktop creates a buffer, this holds the file name to visit."
+  :type '(choice file (const nil))
+  :group 'desktop)
+
+(defcustom desktop-buffer-name nil
+  "When desktop creates a buffer, this holds the desired buffer name."
+  :type '(choice string (const nil))
+  :group 'desktop)
 
 (defvar desktop-buffer-misc nil
   "When desktop creates a buffer, this holds a list of misc info.
 It is used by the `desktop-buffer-handlers' functions.")
 
+(defcustom desktop-buffer-misc-functions
+  '(desktop-buffer-info-misc-data
+    desktop-buffer-dired-misc-data)
+  "*Functions used to determine auxiliary information for a buffer.
+These functions are called in order, with no arguments.  If a function
+returns non-nil, its value is saved along with the desktop buffer for
+which it was called; no further functions will be called.
+
+Later, when desktop.el restores the buffers it has saved, each of the
+`desktop-buffer-handlers' functions will have access to a buffer local
+variable, named `desktop-buffer-misc', whose value is what the
+\"misc\" function returned previously."
+  :type '(repeat function)
+  :group 'desktop)
+
 (defcustom desktop-buffer-handlers
   '(desktop-buffer-dired
     desktop-buffer-rmail
@@ -208,14 +259,31 @@ If the function returns t then the buffer is considered created."
   :type '(repeat function)
   :group 'desktop)
 
+(put 'desktop-buffer-handlers 'risky-local-variable t)
+
 (defvar desktop-create-buffer-form "(desktop-create-buffer 205"
   "Opening of form for creation of new buffers.")
 
 (defcustom desktop-save-hook nil
-  "Hook run before saving the desktop to allow you to cut history lists and
-the like shorter."
+  "Hook run before desktop saves the state of Emacs.
+This is useful for truncating history lists, for example."
   :type 'hook
   :group 'desktop)
+
+(defcustom desktop-minor-mode-table
+  '((auto-fill-function auto-fill-mode)
+    (vc-mode nil))
+  "Table mapping minor mode variables to minor mode functions.
+Each entry has the form (NAME RESTORE-FUNCTION).
+NAME is the name of the buffer-local variable indicating that the minor
+mode is active.  RESTORE-FUNCTION is the function to activate the minor mode.
+called.  RESTORE-FUNCTION nil means don't try to restore the minor mode.
+Only minor modes for which the name of the buffer-local variable
+and the name of the minor mode function are different have to added to
+this table."
+  :type 'sexp
+  :group 'desktop)
+
 ;; ----------------------------------------------------------------------------
 (defvar desktop-dirname nil
   "The directory in which the current desktop file resides.")
@@ -228,6 +296,7 @@ the like shorter."
 
 (defvar desktop-delay-hook nil
   "Hooks run after all buffers are loaded; intended for internal use.")
+
 ;; ----------------------------------------------------------------------------
 (defun desktop-truncate (l n)
   "Truncate LIST to at most N elements destructively."
@@ -235,7 +304,16 @@ the like shorter."
     (if (consp here)
 	(setcdr here nil))))
 ;; ----------------------------------------------------------------------------
-(defun desktop-clear () "Empty the Desktop."
+(defcustom desktop-clear-preserve-buffers
+  '("*scratch*" "*Warnings*")
+  "*Buffer names that `desktop-clear' should not delete."
+  :type '(repeat string)
+  :group 'desktop)
+
+(defun desktop-clear ()
+  "Empty the Desktop.
+This kills all buffers except for internal ones
+and those listed in `desktop-clear-preserve-buffers'."
   (interactive)
   (setq kill-ring nil
 	kill-ring-yank-pointer nil
@@ -243,10 +321,15 @@ the like shorter."
 	search-ring-yank-pointer nil
 	regexp-search-ring nil
 	regexp-search-ring-yank-pointer nil)
-  (mapcar (function (lambda (x)
-		      ;; XEmacs change
-		      (if (not (equal (buffer-name x) "*Warnings*"))
-			  (kill-buffer x)))) (buffer-list))
+  (let ((buffers (buffer-list)))
+    (while buffers
+      (or (member (buffer-name (car buffers)) desktop-clear-preserve-buffers)
+	  (null (buffer-name (car buffers)))
+	  ;; Don't kill buffers made for internal purposes.
+	  (and (not (equal (buffer-name (car buffers)) ""))
+	       (eq (aref (buffer-name (car buffers)) 0) ?\ ))
+	  (kill-buffer (car buffers)))
+      (setq buffers (cdr buffers))))
   (delete-other-windows))
 ;; ----------------------------------------------------------------------------
 (add-hook 'kill-emacs-hook 'desktop-kill)
@@ -277,7 +360,7 @@ TXT is a string that when read and evaluated yields value.
 QUOTE may be `may' (value may be quoted),
 `must' (values must be quoted), or nil (value may not be quoted)."
   (cond
-   ((or (numberp val) (null val) (eq t val))
+   ((or (numberp val) (characterp val) (null val) (eq t val))
     (cons 'may (prin1-to-string val)))
    ((stringp val)
     (let ((copy (copy-sequence val)))
@@ -377,7 +460,7 @@ which means to truncate VAR's value to at most MAX-SIZE elements
 	  (if (and (integerp size)
 		   (> size 0)
 		   (listp (eval var)))
-	      (desktop-truncate (eval var) size)) 
+	      (desktop-truncate (eval var) size))
 	  (insert "(setq "
 		  (symbol-name var)
 		  " "
@@ -389,16 +472,17 @@ which means to truncate VAR's value to at most MAX-SIZE elements
 FILENAME is the visited file name, BUFNAME is the buffer name, and
 MODE is the major mode."
   (let ((case-fold-search nil))
-    (or (and filename
-	     (not (string-match desktop-buffers-not-to-save bufname))
-	     (not (string-match desktop-files-not-to-save filename)))
-	(and (eq mode 'dired-mode)
-	     (save-excursion
-	       (set-buffer (get-buffer bufname))
-	       (not (string-match desktop-files-not-to-save
-				  default-directory))))
-	(and (null filename)
-	     (memq mode '(Info-mode rmail-mode))))))
+    (and (not (string-match desktop-buffers-not-to-save bufname))
+	 (not (memq mode desktop-modes-not-to-save))
+	 (or (and filename
+		  (not (string-match desktop-files-not-to-save filename)))
+	     (and (eq mode 'dired-mode)
+		  (save-excursion
+		    (set-buffer (get-buffer bufname))
+		    (not (string-match desktop-files-not-to-save
+				       default-directory))))
+	     (and (null filename)
+		  (memq mode desktop-buffer-modes-to-save))))))
 ;; ----------------------------------------------------------------------------
 (defun desktop-save (dirname)
   "Save the Desktop file.  Parameter DIRNAME specifies where to save desktop."
@@ -408,41 +492,45 @@ MODE is the major mode."
     (let ((filename (expand-file-name desktop-basefilename dirname))
 	  (info (nreverse
 		 (mapcar
-		  (function (lambda (b)
-			      (set-buffer b)
-			      (list
-			       (buffer-file-name)
-			       (buffer-name)
-			       major-mode
-			       (list	; list explaining minor modes
-				(not (null auto-fill-function)))
-			       (point)
-			       (list (mark t) ;; mark-active
-				     (not (null (mark)))) ; XEmacs
-			       buffer-read-only
-			       (cond ((eq major-mode 'Info-mode)
-				      (list Info-current-file
-					    Info-current-node))
-				     ((eq major-mode 'dired-mode)
-				      (cons
-				       (expand-file-name dired-directory)
-				       (cdr
-					(nreverse
-					 (mapcar
-					  (function car)
-					  dired-subdir-alist))))))
-			       (let ((locals desktop-locals-to-save)
-				     (loclist (buffer-local-variables))
-				     (ll))
-				 (while locals
-				   (let ((here (assq (car locals) loclist)))
-				     (if here
-					 (setq ll (cons here ll))
-				       (if (member (car locals) loclist)
-					   (setq ll (cons (car locals) ll)))))
-				   (setq locals (cdr locals)))
-				 ll)
-			       )))
+		  (function
+		   (lambda (b)
+                     (set-buffer b)
+                     (list
+                      (buffer-file-name)
+                      (buffer-name)
+                      major-mode
+                      ;; minor modes
+                      (let (ret)
+                        (mapcar
+                         #'(lambda (mim)
+                             (and (boundp mim)
+                                  (symbol-value mim)
+                                  (setq ret
+                                        (cons (let ((special (assq mim desktop-minor-mode-table)))
+                                                (if special
+                                                    (cadr special)
+                                                  mim))
+                                              ret))))
+                         (mapcar #'car minor-mode-alist))
+                        ret)
+                      (point)
+                      (list (mark t) ;; mark-active
+                            (not (null (mark)))) ; XEmacs
+                      buffer-read-only
+                      (run-hook-with-args-until-success
+                       'desktop-buffer-misc-functions)
+                      (let ((locals desktop-locals-to-save)
+                            (loclist (buffer-local-variables))
+                            (ll))
+                        (while locals
+                          (let ((here (assq (car locals) loclist)))
+                            (if here
+                                (setq ll (cons here ll))
+                              (if (member (car locals) loclist)
+                                  (setq ll (cons (car locals) ll)))))
+                          (setq locals (cdr locals)))
+                        ll)
+                      )))
 		  (buffer-list))))
 	  (buf (get-buffer-create "*desktop*")))
       (set-buffer buf)
@@ -486,6 +574,7 @@ MODE is the major mode."
 	(if (file-exists-p filename)
 	    (delete-file filename)))))
 ;; ----------------------------------------------------------------------------
+;;;###autoload
 (defun desktop-read ()
   "Read the Desktop file and the files it specifies.
 This is a no-op when Emacs is running in batch mode."
@@ -508,6 +597,7 @@ This is a no-op when Emacs is running in batch mode."
 	    (message "Desktop loaded."))
 	(desktop-clear)))))
 ;; ----------------------------------------------------------------------------
+;;;###autoload
 (defun desktop-load-default ()
   "Load the `default' start-up library manually.
 Also inhibit further loading of it.  Call this from your `.emacs' file
@@ -519,18 +609,38 @@ to provide correct modes for autoloaded files."
 ;; ----------------------------------------------------------------------------
 ;; Note: the following functions use the dynamic variable binding in Lisp.
 ;;
+(defun desktop-buffer-info-misc-data ()
+  (if (eq major-mode 'Info-mode)
+      (list Info-current-file
+            Info-current-node)))
+
+(defun desktop-buffer-dired-misc-data ()
+  (if (eq major-mode 'dired-mode)
+      (cons
+       (expand-file-name dired-directory)
+       (cdr
+        (nreverse
+         (mapcar
+          (function car)
+          dired-subdir-alist))))))
+
 (defun desktop-buffer-info () "Load an info file."
   (if (eq 'Info-mode desktop-buffer-major-mode)
       (progn
-	(require 'info)
-	(Info-find-node (nth 0 desktop-buffer-misc)
-			(nth 1 desktop-buffer-misc))
-	t)))
+	(let ((first (nth 0 desktop-buffer-misc))
+	      (second (nth 1 desktop-buffer-misc)))
+	(when (and first second)
+	  (require 'info)
+	  (Info-find-node first second)
+	  (current-buffer))))))
 ;; ----------------------------------------------------------------------------
 (defun desktop-buffer-rmail () "Load an RMAIL file."
   (if (eq 'rmail-mode desktop-buffer-major-mode)
       (condition-case error
-	  (progn (rmail-input desktop-buffer-file-name) t)
+	  (progn (rmail-input desktop-buffer-file-name)
+                         (if (eq major-mode 'rmail-mode)
+                             (current-buffer)
+                           rmail-buffer))
 	(file-locked
 	 (kill-buffer (current-buffer))
 	 'ignored))))
@@ -540,16 +650,16 @@ to provide correct modes for autoloaded files."
       (progn
 	(require 'mh-e)
 	(mh-find-path)
-	(mh-visit-folder desktop-buffer-name)
-	t)))
+        (mh-visit-folder desktop-buffer-name)
+	(current-buffer))))
 ;; ----------------------------------------------------------------------------
 (defun desktop-buffer-dired () "Load a directory using dired."
   (if (eq 'dired-mode desktop-buffer-major-mode)
       (if (file-directory-p (file-name-directory (car desktop-buffer-misc)))
 	  (progn
-	    (dired (car desktop-buffer-misc))
-	    (mapcar 'dired-insert-subdir (cdr desktop-buffer-misc))
-	    t)
+            (dired (car desktop-buffer-misc))
+	    (mapcar 'dired-maybe-insert-subdir (cdr desktop-buffer-misc))
+	    (current-buffer))
 	(message "Directory %s no longer exists." (car desktop-buffer-misc))
 	(sit-for 1)
 	'ignored)))
@@ -561,11 +671,18 @@ to provide correct modes for autoloaded files."
 		   (y-or-n-p (format
 			      "File \"%s\" no longer exists. Re-create? "
 			      desktop-buffer-file-name))))
-	  (progn (find-file desktop-buffer-file-name) t)
+	  (let ((buf (find-file-noselect desktop-buffer-file-name)))
+	    (condition-case nil
+		(switch-to-buffer buf)
+	      (error (pop-to-buffer buf))))
 	'ignored)))
 ;; ----------------------------------------------------------------------------
 ;; Create a buffer, load its file, set is mode, ...;  called from Desktop file
 ;; only.
+
+(defvar desktop-last-buffer nil
+  "Last buffer read.  Dynamically bound in `desktop-read'.")
+
 (defun desktop-create-buffer (ver desktop-buffer-file-name desktop-buffer-name
 				  desktop-buffer-major-mode
 				  mim pt mk ro desktop-buffer-misc
@@ -577,38 +694,42 @@ to provide correct modes for autoloaded files."
       (setq handler (car hlist))
       (setq result (funcall handler))
       (setq hlist (cdr hlist)))
-    (if (eq result t)
-	(progn
-	  (if (not (equal (buffer-name) desktop-buffer-name))
-	      (rename-buffer desktop-buffer-name))
-	  (auto-fill-mode (if (nth 0 mim) 1 0))
-	  (goto-char pt)
-	  (if (consp mk)
+    (when (bufferp result)
+      (setq desktop-last-buffer result)
+      (set-buffer result)
+      (if (not (equal (buffer-name) desktop-buffer-name))
+	  (rename-buffer desktop-buffer-name))
+      ;; minor modes
+      (cond ((equal '(t) mim)   (auto-fill-mode 1))	; backwards compatible
+	    ((equal '(nil) mim) (auto-fill-mode 0))
+	    (t (mapcar #'(lambda (minor-mode)
+			   (when (functionp minor-mode)
+			     (funcall minor-mode 1)))
+		       mim)))
+      (goto-char pt)
+      (if (consp mk)
+	  (progn
+	    (set-mark (car mk))
+	    (if (car (cdr mk)) (zmacs-activate-region)))
+	(set-mark mk))
+      ;; Never override file system if the file really is read-only marked.
+      (if ro (setq buffer-read-only ro))
+      (while locals
+	(let ((this (car locals)))
+	  (if (consp this)
+	      ;; an entry of this form `(symbol . value)'
 	      (progn
-		(set-mark (car mk))
-;;		(setq mark-active (car (cdr mk))))
-                (if (car (cdr mk)) (zmacs-activate-region))) ; XEmacs
-	    (set-mark mk))
-	  ;; Never override file system if the file really is read-only marked.
-	  (if ro (setq buffer-read-only ro))
-	  (while locals
-	    (let ((this (car locals)))
-	      (if (consp this)
-		  ;; an entry of this form `(symbol . value)'
-		  (progn
-		    (make-local-variable (car this))
-		    (set (car this) (cdr this)))
-		;; an entry of the form `symbol'
-		(make-local-variable this)
-		(makunbound this)))
-	    (setq locals (cdr locals)))
-	  ))))
+		(make-local-variable (car this))
+		(set (car this) (cdr this)))
+	    ;; an entry of the form `symbol'
+	    (make-local-variable this)
+	    (makunbound this)))
+	(setq locals (cdr locals))))))
 
 ;; Backward compatibility -- update parameters to 205 standards.
 (defun desktop-buffer (desktop-buffer-file-name desktop-buffer-name
-						desktop-buffer-major-mode
-						mim pt mk ro tl fc cfs cr
-						desktop-buffer-misc)
+		       desktop-buffer-major-mode
+		       mim pt mk ro tl fc cfs cr desktop-buffer-misc)
   (desktop-create-buffer 205 desktop-buffer-file-name desktop-buffer-name
 			 desktop-buffer-major-mode (cdr mim) pt mk ro
 			 desktop-buffer-misc
@@ -618,6 +739,16 @@ to provide correct modes for autoloaded files."
 			       (cons 'case-replace cr)
 			       (cons 'overwrite-mode (car mim)))))
 ;; ----------------------------------------------------------------------------
+
+;; If the user set desktop-enable to t with Custom,
+;; do the rest of what it takes to use desktop,
+;; but do it after finishing loading the init file.
+(add-hook 'after-init-hook
+	  '(lambda ()
+	     (when desktop-enable
+	       (desktop-load-default)
+	       (desktop-read))))
+
 (provide 'desktop)
 
-;; desktop.el ends here.
+;;; desktop.el ends here
