@@ -1,7 +1,8 @@
 ;;; recent-files.el --- Maintain menu of recently opened files.
-;;; $Header: /usr/CVSroot/XEmacs/xemacs-packages/oa/edit-utils/recent-files.el,v 1.4 2000/10/06 09:08:01 youngs Exp $
+;;; $Header: /pack/xemacscvs/XEmacs/packages/xemacs-packages/edit-utils/recent-files.el,v 1.7 2002/05/11 05:41:23 youngs Exp $
 ;;;
 ;;; Copyright (C) 1994, 1995 Juergen Nickelsen <nickel@cs.tu-berlin.de>
+;;; Copyright (C) 2002 Ben Wing.
 ;;;
 ;; Keywords: menu, file
 
@@ -159,7 +160,8 @@
 ;;; `recent-files-non-permanent-submenu'
 ;;;    If this variable is non-nil, the non-permanent entries are put
 ;;;    into a separate submenu of the recent-files menu. The default
-;;;    is nil. (You can set both `recent-files-permanent-submenu' and
+;;;    is currently t, but probably should be nil, and we may change it
+;;;    back. (You can set both `recent-files-permanent-submenu' and
 ;;;    `recent-files-non-permanent-submenu' to t to have both lists in
 ;;;    separate submenus.)
 ;;;
@@ -371,6 +373,8 @@ This definition is evaluated in a context where `filename' holds the file
 name of the current buffer and `lastpart' holds the last component of
 `filename'.")
 
+(defvar recent-files-menu-omit-chars-list '(?n ?m ?e ?d ?s ?k))
+
 
 (defconst recent-files-save-file-header
   ";; This file is generated by recent-files.
@@ -394,6 +398,7 @@ If the buffer has changed, the menu must be rebuilt.")
 
 ;;; Module initialization
 
+;;;###autoload
 (defun recent-files-initialize ()
   "Initialize the recent-files menu."
   (interactive)
@@ -416,7 +421,7 @@ If the buffer has changed, the menu must be rebuilt.")
   "Return a string identifying the current version of recent-files.
 If called interactively, show it in the echo area."
   (interactive)
-  (let ((version "$Header: /usr/CVSroot/XEmacs/xemacs-packages/oa/edit-utils/recent-files.el,v 1.4 2000/10/06 09:08:01 youngs Exp $"))
+  (let ((version "$Header: /pack/xemacscvs/XEmacs/packages/xemacs-packages/edit-utils/recent-files.el,v 1.7 2002/05/11 05:41:23 youngs Exp $"))
     (if (interactive-p)
 	(message version)
       version)))
@@ -474,6 +479,15 @@ the file."
 
 ;;; Construct the menu
 
+(defun recent-files-maybe-generate-accelerator-spec (list)
+  "Add auto-generated accelerators, if such capability exists.
+This is a basic wrapper around `submenu-generate-accelerator-spec'; if it
+exists, we just call it; otherwise, we do nothing."
+  (if (fboundp 'submenu-generate-accelerator-spec)
+      (submenu-generate-accelerator-spec list
+					 recent-files-menu-omit-chars-list)
+    list))
+
 (defun recent-files-update-menu ()
   "Update the recent-files menu from the recent-files-list."
   (if (or recent-files-list-changed-p
@@ -498,40 +512,58 @@ the file."
 		(null recent-files-permanent-first)
 		(eq t recent-files-permanent-first))
 	    (progn 
-	      (setq permanent (recent-files-make-file-menu-entries
-			       recent-files-list
-			       (function recent-files-filter-permanent)))
-	      (setq non-permanent (recent-files-make-file-menu-entries
-				   recent-files-list
-				   (function
-				    recent-files-filter-non-permanent))))
-	  (setq all (recent-files-make-file-menu-entries
-		     recent-files-list
-		     (function (lambda (l) l)))))
+	      (setq permanent (recent-files-maybe-generate-accelerator-spec
+			       (recent-files-make-file-menu-entries
+				recent-files-list
+				(function recent-files-filter-permanent))))
+	      (setq non-permanent (recent-files-maybe-generate-accelerator-spec
+				   (recent-files-make-file-menu-entries
+				    recent-files-list
+				    (function
+				     recent-files-filter-non-permanent)))))
+	  (setq all (recent-files-maybe-generate-accelerator-spec
+		     (recent-files-make-file-menu-entries
+		      recent-files-list
+		      (function (lambda (l) l))))))
 	(if recent-files-permanent-submenu
-	    (setq permanent (list (cons "Permanent entries..." permanent))))
+	    (setq permanent
+		  (list (cons "Permanent entries..."
+			      permanent))))
 	(if recent-files-non-permanent-submenu
-	    (setq non-permanent (list (cons "Non-permanent entries..."
-					    non-permanent))))
+	    (setq non-permanent
+		  (list (cons "Non-permanent entries..."
+			      non-permanent))))
     ;;; ... and now even uglier.
-	(add-menu recent-files-menu-path recent-files-menu-title
-		  (nconc
-		   (if recent-files-actions-on-top
-		       (append action-menu-entries (list "-----")))
-		   (if (or recent-files-permanent-submenu
-			   recent-files-non-permanent-submenu)
-		       (if recent-files-permanent-first
-			   (nconc permanent non-permanent)
-			 (nconc non-permanent permanent))
-		     (cond ((eq t recent-files-permanent-first)
-			    (nconc permanent non-permanent))
-			   ((null recent-files-permanent-first)
-			    (nconc non-permanent permanent))
-			   (t all)))
-		   (if (not recent-files-actions-on-top)
-		       (cons "-----"
-			     action-menu-entries)))
-		  recent-files-add-menu-before)
+	(let ((menu (nconc
+		     (if recent-files-actions-on-top
+			 (append action-menu-entries (list "-----")))
+		     (if (or recent-files-permanent-submenu
+			     recent-files-non-permanent-submenu)
+			 (if recent-files-permanent-first
+			     (nconc permanent non-permanent)
+			   (nconc non-permanent permanent))
+		       ;; in this case, neither permanent nor non-permanent
+		       ;; files are in submenus.  we need to decide whether
+		       ;; to separate them or combine them, and in the
+		       ;; former case, which one goes first.  also, we
+		       ;; regenerate the accelerator specs to be continuous
+		       ;; across all of them. (it's easier to do it this
+		       ;; way than try to carefully avoid generating them
+		       ;; twice, and totally safe, since the
+		       ;; spec-generating code knows how to remove existing
+		       ;; auto-generated specs before adding new ones.)
+		       (cond ((eq t recent-files-permanent-first)
+			      (recent-files-maybe-generate-accelerator-spec
+			       (nconc permanent non-permanent)))
+			     ((null recent-files-permanent-first)
+			      (recent-files-maybe-generate-accelerator-spec
+			       (nconc non-permanent permanent)))
+			     (t all)))
+		     (if (not recent-files-actions-on-top)
+			 (cons "-----"
+			       action-menu-entries)))))
+	  (add-menu recent-files-menu-path recent-files-menu-title
+		    menu recent-files-add-menu-before))
 	(setq recent-files-list-changed-p nil)
 	(setq recent-files-last-buffer (current-buffer))))
   nil)
@@ -588,14 +620,15 @@ the file."
 
 (defun recent-files-entry-to-menu-entry (entry)
   "Build a menu entry from an entry in `recent-files-list'."
-  (vector (concat (if (cdr entry)
-		      "* "
-		    "  ")
-		  (if recent-files-use-full-names
-		      (recent-files-replace-filenames (car entry))
-		    (recent-files-last-part-of-name (car entry))))
-	  (list recent-files-find-file-command (car entry))
-	  t))
+  (let ((prepend (if (cdr entry)
+                     (if recent-files-permanent-submenu "" "* ")
+		   (if recent-files-non-permanent-submenu "" "  "))))
+    (vector (concat prepend
+                    (if recent-files-use-full-names
+                        (recent-files-replace-filenames (car entry))
+		      (recent-files-last-part-of-name (car entry))))
+            (list recent-files-find-file-command (car entry))
+            t)))
 
 
 (defun recent-files-replace-filenames (filename)
